# mypy: disable-error-code="index,call-arg,union-attr"
import asyncio
import logging
from typing import Any

import httpx
import msgspec

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GrafanaMeta, Link, Resource
from lueur.platform.grafana.client import Client
from lueur.rules import iter_resource

__all__ = ["explore_team", "expand_links"]
logger = logging.getLogger("lueur.lib")


async def explore_team(
    stack_url: str,
    token: str,
) -> list[Resource]:
    resources = []

    tasks: list[asyncio.Task] = []

    async with Client(stack_url, token) as c:
        teams = await explore_teams(c)
        resources.extend(teams)

        async with asyncio.TaskGroup() as tg:
            for team in teams:
                tasks.append(
                    tg.create_task(explore_team_members(c, team.struct["uid"]))
                )

    for task in tasks:
        result = task.result()
        if result is None:
            continue
        resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_teams(c: httpx.AsyncClient) -> list[Resource]:
    response = await c.get("/api/teams/search")

    if response.status_code == 403:
        logger.warning("Grafana API authentication failed")
        return []

    teams = msgspec.json.decode(response.content)

    if "teams" not in teams:
        logger.warning(f"No Grafana teams group found: {teams}")
        return []

    results = []
    for team in teams["teams"]:
        results.append(
            Resource(
                id=make_id(f"{team['id']}-{team['orgId']}"),
                meta=GrafanaMeta(
                    name=team["name"],
                    display=team["name"],
                    kind="team",
                    platform="grafana",
                    category="observability",
                ),
                struct=team,
            )
        )

    return results


async def explore_team_members(
    c: httpx.AsyncClient, team_id: int
) -> list[Resource]:
    response = await c.get(f"/api/teams/{team_id}/members")

    if response.status_code == 403:
        logger.warning("Grafana API authentication failed")
        return []

    members = msgspec.json.decode(response.content)

    results = []
    for member in members:
        results.append(
            Resource(
                id=make_id(f"{member['userId']}-{team_id}"),
                meta=GrafanaMeta(
                    name=member["email"],
                    display=member["email"],
                    kind="team-member",
                    platform="grafana",
                    category="observability",
                ),
                struct=member,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for team_id in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='team' && @.meta.platform=='grafana'].struct.uid",  # noqa: E501
    ):
        r_id = team_id.parent.parent.obj["id"]  # type: ignore
        uid = team_id.value

        p = f"$.resources[?@.meta.kind=='team-member' && @.struct.teamUID=='{uid}']"  # noqa E501
        for member in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="team-member",
                    path=member.path,
                    pointer=str(member.pointer()),
                ),
            )

    for team_id in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='team-member' && @.meta.platform=='grafana'].struct.teamUID",  # noqa: E501
    ):
        r_id = team_id.parent.parent.obj["id"]  # type: ignore
        uid = team_id.value

        p = f"$.resources[?@.meta.kind=='team' && @.struct.uid=='{uid}']"
        for team in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="team",
                    path=team.path,
                    pointer=str(team.pointer()),
                ),
            )



# mypy: disable-error-code="call-arg,index"
import logging
from typing import Any

import msgspec
from kubernetes import client

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, K8SMeta, Link, Resource
from lueur.platform.k8s.client import AsyncClient, Client
from lueur.resource import filter_out_keys
from lueur.rules import iter_resource

__all__ = ["explore_deployment"]
logger = logging.getLogger("lueur.lib")


async def explore_deployment(
    credentials: dict[str, Any] | None,
) -> list[Resource]:
    resources = []

    async with Client(client.AppsV1Api, credentials=credentials) as c:
        pods = await explore_deployments(c)
        resources.extend(pods)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_deployments(c: AsyncClient) -> list[Resource]:
    f = "list_deployment_for_all_namespaces"
    response = await c.execute(f)

    if response.status == 403:
        logger.warning("Kubernetes API server failed authentication")
        return []

    if response.status == 404:
        logger.warning(f"Kubernetes API server '{f}' not found")
        return []

    deployments = msgspec.json.decode(response.data)

    if "items" not in deployments:
        logger.warning(f"No deployments found: {deployments}")
        return []

    results = []
    for deployment in deployments["items"]:
        meta = deployment["metadata"]
        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=meta["name"],
                    display=meta["name"],
                    kind="deployment",
                    platform="k8s",
                    ns=meta["namespace"],
                    category="compute",
                ),
                struct=filter_out_keys(
                    deployment, keys=[["metadata", "managedFields"]]
                ),
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for deployment in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='deployment' && @.meta.platform=='k8s'].meta.name",  # noqa: E501
    ):
        r_id = deployment.parent.parent.obj["id"]  # type: ignore
        name: str = deployment.value  # type: ignore
        ns = deployment.parent.parent.obj["meta"]["ns"]  # type: ignore

        add_rs_links(d, serialized, r_id, name, ns)


def add_rs_links(
    d: Discovery,
    serialized: dict[str, Any],
    deployment_resource_id: str,
    deployment_name: str,
    deployment_ns: str,
) -> None:
    p = f"$.resources[?@.meta.kind=='replicaset'].struct.metadata.ownerReferences[?@.kind=='Deployment' && @.name=='{deployment_name}']"  # noqa E501
    for ownerRef in iter_resource(serialized, p):
        rs = ownerRef.parent.parent.parent.parent  # type: ignore
        add_link(
            d,
            deployment_resource_id,
            Link(
                direction="out",
                kind="replicaset",
                id=rs.obj["id"],  # type: ignore
                path=rs.path,  # type: ignore
                pointer=str(rs.pointer()),  # type: ignore
            ),
        )

        rs_name = rs.obj["meta"]["name"]  # type: ignore

        add_pod_links(
            d, serialized, deployment_resource_id, rs_name, deployment_ns
        )


def add_pod_links(
    d: Discovery,
    serialized: dict[str, Any],
    deployment_resource_id: str,
    rs_name: str,
    rs_ns: str,
) -> None:
    p = f"$.resources[?@.meta.kind=='pod'].struct.metadata.ownerReferences[?@.kind=='ReplicaSet' && @.name=='{rs_name}']"  # noqa E501
    for ownerRef in iter_resource(serialized, p):
        po = ownerRef.parent.parent.parent.parent  # type: ignore
        add_link(
            d,
            deployment_resource_id,
            Link(
                direction="out",
                kind="pod",
                id=po.obj["id"],  # type: ignore
                path=po.path,  # type: ignore
                pointer=str(po.pointer()),  # type: ignore
            ),
        )

        node_ip = po.obj["struct"]["status"]["hostIP"]  # type: ignore
        pod_id = po.obj["id"]  # type: ignore

        add_node_links(d, serialized, deployment_resource_id, node_ip)
        add_service_links(d, serialized, deployment_resource_id, pod_id, rs_ns)


def add_node_links(
    d: Discovery,
    serialized: dict[str, Any],
    deployment_resource_id: str,
    node_ip: str,
) -> None:
    p = f"$.resources[?@.meta.kind=='node'].struct.status.addresses[?@address=='{node_ip}']"  # noqa E501
    for addr in iter_resource(serialized, p):
        node = addr.parent.parent.parent.parent  # type: ignore
        add_link(
            d,
            deployment_resource_id,
            Link(
                direction="out",
                kind="node",
                id=node.obj["id"],  # type: ignore
                path=node.path,  # type: ignore
                pointer=str(node.pointer()),  # type: ignore
            ),
        )


def add_service_links(
    d: Discovery,
    serialized: dict[str, Any],
    deployment_resource_id: str,
    pod_id: str,
    pod_ns: str,
) -> None:
    for svc in iter_resource(
        serialized,
        f"$.resources[?@.meta.kind=='service' && @.meta.platform=='k8s' && @.meta.ns=='{pod_ns}']",  # noqa: E501
    ):
        selectors = svc.obj["struct"]["spec"].get("selector")  # type: ignore
        if not selectors:
            continue

        labels = " && ".join(
            [
                f"@.struct.metadata.labels.['{k}']=='{v}'"
                for k, v in selectors.items()
            ]
        )  # noqa: E501

        p = f"$.resources[?@.id=='{pod_id}' && @.meta.kind=='pod' && @.meta.ns=='{pod_ns}' && {labels}]"  # noqa: E501
        for pod in iter_resource(serialized, p):
            # we found a match between our current pod and a service
            add_link(
                d,
                deployment_resource_id,
                Link(
                    direction="out",
                    kind="service",
                    id=svc.obj["id"],  # type: ignore
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )

            svc_name = svc.obj["meta"]["name"]

            p = f"$.resources[?@.meta.kind=='httproute'].struct.spec.rules.*.backendRefs[?@.kind=='Service' && @.name=='{svc_name}']"  # noqa: E501
            for httproute in iter_resource(serialized, p):
                httproute = httproute.parent.parent.parent.parent.parent.parent  # type: ignore

                add_link(
                    d,
                    deployment_resource_id,
                    Link(
                        direction="out",
                        kind="httproute",
                        id=httproute.obj["id"],  # type: ignore
                        path=httproute.path,
                        pointer=str(httproute.pointer()),
                    ),
                )

                for gw in httproute.obj["struct"]["spec"]["parentRefs"]:
                    if gw["kind"] != "Gateway":
                        continue

                    gw_name = gw["name"]
                    p = f"$.resources[?@.meta.kind=='gateway' && @.meta.display=='{gw_name}']"  # noqa: E501
                    for gw in iter_resource(serialized, p):
                        add_link(
                            d,
                            deployment_resource_id,
                            Link(
                                direction="out",
                                kind="gateway",
                                id=gw.obj["id"],
                                path=gw.path,
                                pointer=str(gw.pointer()),
                            ),
                        )

            break



from typing import Any, Iterable

import jsonpath

__all__ = ["iter_resource", "match_path", "filter_out_keys"]


def iter_resource(
    discovery: dict[str, Any], path: str
) -> Iterable[jsonpath.JSONPathMatch]:
    yield from jsonpath.finditer(path, discovery)


def match_path(
    discovery: dict[str, Any], path: str
) -> jsonpath.JSONPathMatch | None:
    return jsonpath.match(path, discovery)


def filter_out_keys(
    struct: dict[str, Any], keys: list[str | list[str]]
) -> dict[str, Any]:
    for k in keys:
        if isinstance(k, dict):
            struct.pop(k, None)
        elif isinstance(k, list):
            current = struct
            for i in k[:-1]:
                current = struct.get(i)  # type: ignore
                if current is None:
                    break

            if current:
                current.pop(k[-1], None)

    return struct



# mypy: disable-error-code="index,call-arg,union-attr"
import logging
from typing import Any

import httpx
import msgspec

from lueur.make_id import make_id
from lueur.models import Discovery, GrafanaMeta, Resource
from lueur.platform.grafana.client import Client

__all__ = ["explore_incident", "expand_links"]
logger = logging.getLogger("lueur.lib")


async def explore_incident(
    stack_url: str,
    token: str,
) -> list[Resource]:
    resources = []

    async with Client(stack_url, token) as c:
        incidents = await explore_incidents(c)
        resources.extend(incidents)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_incidents(c: httpx.AsyncClient) -> list[Resource]:
    response = await c.post(
        "/api/plugins/grafana-incident-app/resources/api/v1/IncidentsService.QueryIncidentPreviews",  # noqa: E501
        json={
            "cursor": {},
            "includeCustomFieldValues": True,
            "includeMembershipPreview": True,
            "query": {
                "limit": 30,
                "orderDirection": "ASC",
                "orderField": "createdTime",
            },
        },
    )

    if response.status_code == 403:
        logger.warning("Grafana API authentication failed")
        return []

    incidents = msgspec.json.decode(response.content)

    if "incidentPreviews" not in incidents:
        logger.warning(f"No Grafana incidents found: {incidents}")
        return []

    results = []
    for incident in incidents["incidentPreviews"]:
        results.append(
            Resource(
                id=make_id(
                    f"{incident['incidentID']}-{incident['createdTime']}"
                ),
                meta=GrafanaMeta(
                    name=incident["slug"],
                    display=incident["title"],
                    kind="incident",
                    platform="grafana",
                    category="observability",
                ),
                struct=incident,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    pass



# mypy: disable-error-code="index,call-arg,union-attr"
import logging
from typing import Any

import httpx
import msgspec

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GrafanaMeta, Link, Resource
from lueur.platform.grafana.client import Client
from lueur.rules import iter_resource

__all__ = ["explore_slo", "expand_links"]
logger = logging.getLogger("lueur.lib")


async def explore_slo(
    stack_url: str,
    token: str,
) -> list[Resource]:
    resources = []

    async with Client(stack_url, token) as c:
        slos = await explore_slos(c)
        resources.extend(slos)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_slos(c: httpx.AsyncClient) -> list[Resource]:
    response = await c.get("/api/plugins/grafana-slo-app/resources/v1/slo")

    if response.status_code == 403:
        logger.warning("Grafana API authentication failed")
        return []

    slos = msgspec.json.decode(response.content)

    if "slos" not in slos:
        logger.warning(f"No Grafana SLO found: {slos}")
        return []

    results = []
    for slo in slos["slos"]:
        results.append(
            Resource(
                id=make_id(slo["uuid"]),
                meta=GrafanaMeta(
                    name=slo["name"],
                    display=slo["name"],
                    kind="slo",
                    platform="grafana",
                    category="observability",
                ),
                struct=slo,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for team_name in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='slo' && @.meta.platform=='grafana'].struct.labels[?@.key=='team_name'].value",  # noqa: E501
    ):
        r_id = team_name.parent.parent.parent.parent.obj["id"]  # type: ignore
        name = team_name.value

        p = f"$.resources[?@.meta.kind=='team' && @.meta.platform=='grafana' && @.meta.name=='{name}']"  # noqa: E501
        for team in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="team",
                    path=team.path,
                    pointer=str(team.pointer()),
                    id=team.obj["id"],
                ),
            )

    for slo in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='slo' && @.meta.platform=='grafana']",  # noqa: E501
    ):
        r_id = slo.obj["id"]  # type: ignore
        slo = slo.value  # type: ignore

        p = f"$.resources[?@.meta.kind=='team' && @.meta.platform=='grafana' && @.meta.name=='{name}']"  # noqa: E501
        for team in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="team",
                    path=team.path,
                    pointer=str(team.pointer()),
                    id=team.obj["id"],
                ),
            )
